import logging
from typing import Union, Dict, Type, Any

log = logging.getLogger(__name__)


class ArsenicError(Exception):
    pass


class SessionStartError(ArsenicError):
    def __init__(self, error, message, response):
        self.error = error
        self.message = message
        self.response = response
        super().__init__(f"{error}: {message}")


class OperationNotSupported(ArsenicError):
    pass


class WebdriverError(ArsenicError):
    def __init__(self, message, screen, stacktrace):
        self.message = message
        self.screen = screen
        self.stacktrace = stacktrace
        super().__init__(message)


class UnknownArsenicError(ArsenicError):
    pass


class ArsenicTimeout(ArsenicError):
    pass


CODES: Dict[Union[str, int], Type[WebdriverError]] = {}


def get(error_code: Union[str, int]) -> Type[WebdriverError]:
    return CODES.get(error_code, UnknownArsenicError)


def create(error_name: str, *error_codes: int) -> Type[WebdriverError]:
    name = "".join(bit.capitalize() for bit in error_name.split(" "))
    cls: Type[WebdriverError] = type(name, (WebdriverError,), {})
    CODES[error_name] = cls
    for code in error_codes:
        CODES[code] = cls
    return cls


NoSuchElement = create("no such element", 7)
NoSuchFrame = create("no such frame", 8)
UnknownCommand = create("unknown command", 9)
StaleElementReference = create("stale element reference", 10)
ElementNotVisible = create("element not visible", 11)
InvalidElementState = create("invalid element state", 12)
UnknownError = create("unknown error", 13)
ElementNotInteractable = create("element not interactable")
ElementIsNotSelectable = create("element is not selectable", 15)
JavascriptError = create("javascript error", 17)
Timeout = create("timeout", 21)
NoSuchWindow = create("no such window", 23)
InvalidCookieDomain = create("invalid cookie domain", 24)
UnableToSetCookie = create("unable to set cookie", 25)
UnexpectedAlertOpen = create("unexpected alert open", 26)
NoSuchAlert = create("no such alert", 27)
ScriptTimeout = create("script timeout", 28)
InvalidElementCoordinates = create("invalid element coordinates", 29)
IMENotAvailable = create("ime not available", 30)
IMEEngineActivationFailed = create("ime engine activation failed", 31)
InvalidSelector = create("invalid selector", 32)
MoveTargetOutOfBounds = create("move target out of bounds", 34)

# exceptions that may indicate errors within arsenic
ERROR_CLASSES = (UnknownArsenicError, UnknownCommand, UnknownError)


def raise_exception(data: Dict[str, Any], status: int):
    error = None
    if "status" in data:
        error = data["status"]
    elif "error" in data:
        error = data["error"]
    elif "state" in data:
        error = data["state"]

    if "value" in data and isinstance(data["value"], dict):
        data = data["value"]
    if error is None and "error" in data:
        error = data["error"]
    message = data.get("message", None)
    stacktrace = data.get("stacktrace", None)
    screen = data.get("screen", None)
    exception_class = get(error)
    if exception_class in ERROR_CLASSES:
        log.error(
            f"Error: {exception_class} - {message} (status={status})\n"
            f"Stacktrace: {stacktrace}\nData: {data}"
        )
    raise exception_class(message, screen, stacktrace)



import abc
from dataclasses import dataclass
from enum import Enum
from typing import Sequence, Any, Iterator, List, Dict, Optional

from wapiti_arsenic import constants
from wapiti_arsenic.session import Element


@dataclass
class Action:
    source: "Device"
    payload: Dict[str, Any]


class Tick:
    def __init__(self, *actions: Action):
        self.actions: Dict["Device", "Action"] = {
            action.source: action for action in actions
        }

    def __and__(self, other: "Tick") -> "Tick":
        overlap = set(self.actions.keys()) & set(other.actions.keys())
        if overlap:
            raise ValueError(
                f"Devices {overlap} have more than one action in this tick"
            )
        return Tick(*self.actions.values(), *other.actions.values())

    def encode(self, device: "Device") -> Dict[str, Any]:
        if device not in self.actions:
            return {"type": "pause", "duration": 0}

        return self.actions[device].payload


class DeviceType(Enum):
    keyboard = "key"
    pointer = "pointer"


class PointerType(Enum):
    mouse = "mouse"
    pen = "pen"
    touch = "touch"


class Button(Enum):
    left = 0
    middle = 1
    right = 2


class Device(metaclass=abc.ABCMeta):
    type: DeviceType = abc.abstractproperty()

    def __init__(self, device_id: Optional[str] = None):
        self.device_id = device_id

    def info(self, index: int) -> Dict[str, Any]:
        device_id = self.device_id or f"{self.type.value}{index}"
        return {"id": device_id, "type": self.type.value}

    def _tick(self, **payload: Any) -> Tick:
        return Tick(Action(self, payload))

    def pause(self, duration: int) -> Tick:
        return self._tick(type="pause", duration=duration)


class Pointer(Device, metaclass=abc.ABCMeta):
    type = DeviceType.pointer
    pointer_type: PointerType = abc.abstractproperty()

    def info(self, index: int) -> Dict[str, Any]:
        return {
            "parameters": {"pointerType": self.pointer_type.value},
            **super().info(index),
        }

    def move_to(self, element: Element, duration: int = 250) -> Tick:
        return self._tick(
            type="pointerMove",
            duration=duration,
            origin={constants.WEB_ELEMENT: element.id},
            x=0,
            y=0,
        )

    def move_by(self, x: int, y: int, duration: int = 250) -> Tick:
        return self._tick(
            type="pointerMove", duration=duration, origin="pointer", x=x, y=y
        )

    def down(self) -> Tick:
        return self._tick(type="pointerDown", duration=0, button=0)

    def up(self) -> Tick:
        return self._tick(type="pointerUp", duration=0, button=0)


class Mouse(Pointer):
    pointer_type = PointerType.mouse

    def down(self, button: Button = Button.left) -> Tick:
        return self._tick(type="pointerDown", duration=0, button=button.value)

    def up(self, button: Button = Button.left) -> Tick:
        return self._tick(type="pointerUp", duration=0, button=button.value)


class Pen(Pointer):
    pointer_type = PointerType.pen


class Touch(Pointer):
    pointer_type = PointerType.touch


class Keyboard(Device):
    type = DeviceType.keyboard

    def down(self, key: str) -> Tick:
        return self._tick(type="keyDown", value=key)

    def up(self, key: str) -> Tick:
        return self._tick(type="keyUp", value=key)


def gather_devices(ticks: Sequence[Tick]) -> Iterator[Device]:
    found = set()
    for tick in ticks:
        devices = set(tick.actions.keys())
        for device in devices - found:
            yield device
        found.update(devices)


def chain(*ticks: Tick) -> Dict[str, List[Dict[str, Any]]]:
    devices = list(gather_devices(ticks))
    return {
        "actions": [
            {**device.info(index), "actions": [tick.encode(device) for tick in ticks]}
            for index, device in enumerate(devices, start=1)
        ]
    }



from enum import Enum

WEB_ELEMENT = "element-6066-11e4-a52e-4f735466cecf"

STATUS_SUCCESS = 0


class SelectorType(Enum):
    css_selector = "css selector"
    link_text = "link text"
    partial_link_text = "partial link text"
    tag_name = "tag name"
    xpath = "xpath"


class WindowType(Enum):
    tab = "tab"
    window = "window"



import sys

if sys.platform != "win32":

    def check_ie11_environment_cli():
        pass

    def configure_ie11_environment_cli():
        pass

else:
    import ctypes
    import winreg
    from collections import defaultdict
    from ctypes import wintypes

    KEYS = [
        (
            winreg.HKEY_LOCAL_MACHINE,
            r"SOFTWARE\Microsoft\Internet Explorer\Main\FeatureControl\FEATURE_BFCACHE",
            "iexplore.exe",
            0,
            winreg.REG_DWORD,
        ),
        (
            winreg.HKEY_LOCAL_MACHINE,
            r"SOFTWARE\Wow6432Node\Microsoft\Internet Explorer\Main\FeatureControl\FEATURE_BFCACHE",
            "iexplore.exe",
            0,
            winreg.REG_DWORD,
        ),
        (
            winreg.HKEY_CURRENT_USER,
            r"Software\Microsoft\Internet Explorer\Zoom",
            "ZoomFactor",
            100_000,
            winreg.REG_DWORD,
        ),
    ]
    PROTECTION_MODES = [
        (
            winreg.HKEY_CURRENT_USER,
            r"Software\Microsoft\Windows\CurrentVersion\Internet Settings\Lockdown_Zones\0",
            "CurrentLevel",
            winreg.REG_DWORD,
        ),
        (
            winreg.HKEY_CURRENT_USER,
            r"Software\Microsoft\Windows\CurrentVersion\Internet Settings\Lockdown_Zones\1",
            "CurrentLevel",
            winreg.REG_DWORD,
        ),
        (
            winreg.HKEY_CURRENT_USER,
            r"Software\Microsoft\Windows\CurrentVersion\Internet Settings\Lockdown_Zones\2",
            "CurrentLevel",
            winreg.REG_DWORD,
        ),
        (
            winreg.HKEY_CURRENT_USER,
            r"Software\Microsoft\Windows\CurrentVersion\Internet Settings\Lockdown_Zones\3",
            "CurrentLevel",
            winreg.REG_DWORD,
        ),
        (
            winreg.HKEY_CURRENT_USER,
            r"Software\Microsoft\Windows\CurrentVersion\Internet Settings\Lockdown_Zones\4",
            "CurrentLevel",
            winreg.REG_DWORD,
        ),
    ]

    KEY_NAMES = {
        value: key for key, value in vars(winreg).items() if key.startswith("HKEY_")
    }

    TYPE_NAMES = {
        value: key for key, value in vars(winreg).items() if key.startswith("REG_")
    }

    def detect_dpi_setting():
        ctypes.windll.shcore.SetProcessDpiAwareness(2)

        ok = []

        def cb(monitor, dc, rect, data):
            x = wintypes.UINT()
            y = wintypes.UINT()
            ctypes.windll.shcore.GetDpiForMonitor(
                monitor, 0, ctypes.pointer(x), ctypes.pointer(y)
            )
            if (x.value, y.value) != (96, 96):
                print("DPI Scaling does not seem to be set to 100%")
                ok.append(False)
            else:
                ok.append(True)
            return 0

        ctypes.windll.user32.EnumDisplayMonitors(
            0,
            0,
            ctypes.WINFUNCTYPE(
                ctypes.c_uint,
                ctypes.c_ulong,
                ctypes.c_ulong,
                ctypes.POINTER(wintypes.RECT),
                ctypes.c_double,
            )(cb),
            0,
        )
        if not ok:
            print("No Monitors found")
        return ok and all(ok)

    def check_ie11_environment():
        ok = True
        for key, subkey, attribute, value, value_type in KEYS:
            path = f"{KEY_NAMES[key]}\\{subkey}"
            try:
                with winreg.OpenKey(key, subkey) as regkey:
                    try:
                        actual_value, actual_type = winreg.QueryValueEx(
                            regkey, attribute
                        )
                    except OSError:
                        ok = False
                        print(f"Key {path} has no attribute {attribute}")
                    else:
                        if actual_value != value:
                            ok = False
                            print(
                                f"Key {path}\\{attribute} has value {actual_value}, expected {value}"
                            )
                        if actual_type != value_type:
                            ok = False
                            print(
                                f"Key {path}\\{attribute} has value type {TYPE_NAMES[actual_type]}, expected {TYPE_NAMES[value_type]}"
                            )
            except OSError:
                ok = False
                print(f"Key {path} does not exist!")
        modes = defaultdict(list)
        for key, subkey, attribute, value_type in PROTECTION_MODES:
            path = f"{KEY_NAMES[key]}\\{subkey}"
            try:
                with winreg.OpenKey(key, subkey) as regkey:
                    try:
                        actual_value, actual_type = winreg.QueryValueEx(
                            regkey, attribute
                        )
                    except OSError:
                        ok = False
                        print(f"Key {path} has no attribute {attribute}")
                    else:
                        modes[actual_value].append(path)
                        if actual_type != value_type:
                            ok = False
                            print(
                                f"Key {path}\\{attribute} has value type {TYPE_NAMES[actual_type]}, expected {TYPE_NAMES[value_type]}"
                            )
            except OSError:
                ok = False
                print(f"Key {path} does not exist!")
        if len(modes) > 1:
            print("Not all zones have same protected mode setting:")
            for value, zones in modes.items():
                print(f'{", ".join(zones)} have value {value}')
        ok = ok and detect_dpi_setting()
        if ok:
            print("Environment looks okay")
        return ok

    def check_ie11_environment_cli():
        if not check_ie11_environment():
            sys.exit(-1)

    def configure_ie11_environment(protected_mode):
        for key, subkey, attribute, value, value_type in KEYS:
            try:
                regkey = winreg.OpenKey(key, subkey, access=winreg.KEY_SET_VALUE)
            except OSError:
                regkey = winreg.CreateKey(key, subkey)
            try:
                winreg.SetValueEx(regkey, attribute, 0, value_type, value)

            finally:
                winreg.CloseKey(regkey)
        for key, subkey, attribute, value_type in PROTECTION_MODES:
            try:
                regkey = winreg.OpenKey(key, subkey, access=winreg.KEY_SET_VALUE)
            except OSError:
                regkey = winreg.CreateKey(key, subkey)
            try:
                winreg.SetValueEx(regkey, attribute, 0, value_type, protected_mode)
            finally:
                winreg.CloseKey(regkey)

    def configure_ie11_environment_cli():
        configure_ie11_environment(0)



from wapiti_arsenic.session import Session


class Browser:
    defaults = {}
    session_class = Session

    def __init__(self, **overrides):
        self.capabilities = {**self.defaults, **overrides}


class Firefox(Browser):
    defaults = {"browserName": "firefox"}


class Chrome(Browser):
    defaults = {"browserName": "chrome"}


class InternetExplorer(Browser):
    session_class = Session
    defaults = {"browserName": "internet explorer"}


IE = InternetExplorer


class MicrosoftEdge(Browser):
    defaults = {"browserName": "MicrosoftEdge"}


Edge = MicrosoftEdge



import abc
import asyncio
import logging
import os
import subprocess
import sys
from typing import List, TypeVar
from asyncio.subprocess import DEVNULL, PIPE

log = logging.getLogger(__name__)

P = TypeVar("P")


def check_event_loop():
    if sys.platform == "win32" and isinstance(
        asyncio.get_event_loop(), asyncio.SelectorEventLoop
    ):
        raise ValueError(
            "SelectorEventLoop is not supported on Windows, use asyncio.ProactorEventLoop instead."
        )


class BaseSubprocessImpl(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def run_process(self, cmd: List[str]) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    async def start_process(self, cmd: List[str], log_file) -> P:
        pass

    @abc.abstractmethod
    async def stop_process(self, process: P) -> None:
        pass


class AsyncioSubprocessImpl(BaseSubprocessImpl):
    async def run_process(self, cmd: List[str]) -> str:
        check_event_loop()
        process = await asyncio.create_subprocess_exec(
            *cmd, stdout=PIPE, stderr=PIPE, stdin=DEVNULL
        )
        out, err = await process.communicate()
        if process.returncode != 0:
            raise Exception(err)

        return out.decode("utf-8")

    async def start_process(self, cmd: List[str], log_file):
        check_event_loop()
        if log_file is os.devnull:
            log_file = DEVNULL
        return await asyncio.create_subprocess_exec(
            *cmd, stdout=log_file, stderr=log_file, stdin=DEVNULL
        )

    async def stop_process(self, process):
        process.terminate()
        try:
            await asyncio.wait_for(process.communicate(), 1)
        except asyncio.futures.TimeoutError:
            process.kill()
        try:
            await asyncio.wait_for(process.communicate(), 1)
        except asyncio.futures.TimeoutError:
            log.warning(
                "Could not terminate process", extra={"process": process, "impl": self}
            )


class ThreadedSubprocessImpl(BaseSubprocessImpl):
    async def run_process(self, cmd: List[str]):
        return await asyncio.get_event_loop().run_in_executor(
            None, self._run_process, cmd
        )

    def _run_process(self, cmd):
        return subprocess.check_output(
            cmd, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL
        ).decode("utf-8")

    async def start_process(self, cmd: List[str], log_file):
        if log_file is os.devnull:
            log_file = subprocess.DEVNULL
        return await asyncio.get_event_loop().run_in_executor(
            None, self._start_process, cmd, log_file
        )

    def _start_process(self, cmd: List[str], log_file):
        return subprocess.Popen(
            cmd, stdout=log_file, stderr=subprocess.STDOUT, stdin=subprocess.DEVNULL
        )

    async def stop_process(self, process):
        return await asyncio.get_event_loop().run_in_executor(
            None, self._stop_process, process
        )

    def _stop_process(self, process: subprocess.Popen):
        process.terminate()
        try:
            process.communicate(timeout=1)
        except subprocess.TimeoutExpired:
            process.kill()
        try:
            process.communicate(timeout=1)
        except subprocess.TimeoutExpired:
            log.warning(
                "Could not terminate process", extra={"process": process, "impl": self}
            )


def get_subprocess_impl() -> BaseSubprocessImpl:
    if sys.platform == "win32":
        if isinstance(asyncio.get_event_loop(), asyncio.SelectorEventLoop):
            return ThreadedSubprocessImpl()
        return AsyncioSubprocessImpl()
    else:
        return AsyncioSubprocessImpl()



import socket
from dataclasses import dataclass
from typing import Union


def free_port() -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.bind(("0.0.0.0", 0))
        sock.listen(5)
        return sock.getsockname()[1]


def px_to_number(value: str) -> Union[int, float]:
    original = value
    if value.endswith("px"):
        value = value[:-2]
    if value.isdigit():
        return int(value)
    try:
        return float(value)
    except ValueError:
        raise ValueError(f"{original!r} is not an number or <number>px value")


@dataclass
class Rect:
    x: float
    y: float
    width: float
    height: float



from dataclasses import dataclass
from typing import Optional

from wapiti_arsenic.browsers import Browser
from wapiti_arsenic.errors import UnknownError
from wapiti_arsenic.services import Service
from wapiti_arsenic.session import Session


@dataclass
class SessionContext:
    service: Service
    browser: Browser
    bind: str
    session: Optional[Session] = None

    async def __aenter__(self):
        self.session = await start_session(self.service, self.browser, self.bind)
        return self.session

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await stop_session(self.session)


def get_session(service, browser, bind=""):
    return SessionContext(service, browser, bind)


async def start_session(service: Service, browser: Browser, bind=""):
    driver = await service.start()
    return await driver.new_session(browser, bind=bind)


async def stop_session(session: Session):
    try:
        await session.close()
    except UnknownError as e:
        if "marionette" not in str(e):
            raise
    await session.driver.close()
